#!/usr/bin/env python
#
'Browse the local testage in a tree.'
#
#  LGPL - don't steal, don't sue
#

import glob, imp, inspect, os, re, string, sys, time, traceback
import Tkinter, Pmw
from unittest import _TextTestResult, _WritelnDecorator
from unittest import findTestCases, TestSuite
from TreeBox import TreeBox
from getopt import getopt
from Debug import db, inspectObject
from os import path
import piper

tk = Tkinter
iz = inspect


_slash = '/'  #  for internal use only; this is not os.sep
_dot = '.'

#  TODO:  save and restore the window geometries
#         same module name in different folders (just warn!)
#         also trigger a retest when imported modules change (!)
#         put a clicker to the top of the test case in the error list
#         ensure the source locator also works on sub-folders
#         route that pesky Tkinter error traceback into here
#         change the color of tree entries that failed to orange or red?
#         at syntax error time, the system just emits a burst of
#                 irrelevancies, none of which are clickable
#         get the db() system to pump info into the list box
#         hack traceback to put in >full< paths
#         option to censor systemic things from the traceback
#         average method length statistics & rolls-up
#         hourglass the cursor or something during long processes
#         ability to put on the command line a test name or test wild card
#         extract __doc__ from __init__.py files
#         demonstrate extended paths on testee module loading
#         trigger when __init__.py files change (fresh __doc__)
#         show test inheritance
#         at retest time, refresh the progress bar in its brief white state
#         put dot above the other folders
#         visual feedback at test start time (cyan snip or something)
#         get the __init__.py __doc__ working
#         Don't croak everything when someone closes Idle
#         deal when a file disappears

relevantArgs = sys.argv[1:]

USE_KATE = 0
USE_IDLE = 1
USE_VIM  = 2


class CommandLineParser:
    def __init__(self):
        self.myKeywordOpts = None
        self.myFreeOpts = None
        self.myWildCards = []
        self.myFolders = []
        self.myFiles = []
        self.myVerbose = 0
        
    def absorbCommandLine(self, args):
        'also barks at users who want both idle and kate'
        
        opts = getopt(args, "V:i:k:v", ['Verbose', 'idle', 'kate', 'vim'])
        self.myKeywordOpts, self.myFreeOpts = opts
        
        for idx in range(len(self.myKeywordOpts)):
            if self.myKeywordOpts[idx][0] == '--Verbose':
                self.myVerbose = 1
                del self.myKeywordOpts[idx]
                break
        
        if len(self.myKeywordOpts) > 1:
            print "You must choose between Idle, Kate and Vim. We know it's hard."
            sys.exit(1)
            
        self.myWildCards = []
        self.myFolders = []
        self.myFiles = []
        
        for thing in self.myFreeOpts:
            if '*' in thing:
                self.myWildCards.append(thing)
            elif path.isdir(thing):
                self.myFolders.append(thing)
            elif path.isfile(thing):
                self.myFiles.append(thing)
            else:
                print "Argument %r not understood" % thing
                sys.exit(1)
        
        if self.myFiles == [] and self.myFolders == []:
            self.myFolders   = [ '.']
        
        if self.myWildCards == [] and self.myFolders != []:
            self.myWildCards = ['*.py']
            
        return
        
    def getEditor(self):
        for arg, value in self.myKeywordOpts:
            if arg == '--vim' :  return USE_VIM
            if arg == '--idle':  return USE_IDLE
            if arg == '--kate':  return USE_KATE
            
        return USE_IDLE

    def getWildCards(self):  return  [] # TODO croak me self.myWildCards
    def getFolders  (self):  return [] # TODO croak me self.myFolders
    
    def getFiles(self):
        'warning: only call this once, else it wastes time'
        
        self.myFiles = map(path.normpath, self.myFiles)
        map(self.findFilesInFolders, self.myFolders)
        self.myFiles = uneek(self.myFiles)
        self.myFiles.sort()
        return self.myFiles
        
    def getVerbose  (self):  return self.myVerbose

    def findFilesInFolders(self, folder):

        for card in self.myWildCards:
            self.applyWildCard(folder, card)
        
        subs = getList(folder, '*')
       
        for sub in subs:
            if sub != '.' and path.isdir(sub):
                self.findFilesInFolders(sub)
        
        return 

    def applyWildCard(self, folder, card):        
        listage = getList(folder, card)
        init_py = os.sep + '__init__.py'

        for file in listage:
            if file[-len(init_py):] != init_py:         
                file = path.normpath(file)
                self.myFiles.append(file)

        return
        

def getList(folder, card):
    sampler = path.join(folder, card)
    listage = glob.glob(sampler)
    return listage
    
        
def uneek(list):
    val_dict = {}
    for val in list:
        val_dict[val] = ''
    return val_dict.keys()
                   
aCommandLine = CommandLineParser()


class BrowserResult(_TextTestResult):
    def __init__(self, stream):
        _TextTestResult.__init__(self, stream, descriptions = 1, verbosity = 0)

    #  style note - don't use magic members. None all members
    #  in a constructor

        self.myBar = None
        self.myTotal = None
        self.myErrorList = None

    def setUserView(self, bar, lister, total):
        self.myBar = bar
        self.myErrorList = lister
        self.myTotal = total
        self.wazStderr = sys.stderr
        sys.stderr = self

    def addSuccess(self, test):
        self.addEvent()

    def addEvent(self):
        if self.myBar is None:  return

        self.myBar.setProgressFraction(
                        self.getFraction(),
                        self.getColor()
                        )

    def addFailure(self, test, err):
        ''' observe errors go to the console AND the list box.
            Each will provide the user different features '''

        _TextTestResult.addFailure(self, test, err)
        self._printErrorList('FAIL', test, err)
        self.addEvent()

    def _printErrorList(self, flavour, test, err):
        ''' Any similarity between this function and the one in
            unittest.py must be just a shocking coincidence '''

        description = self.getDescription(test)
        complaint = '%s: %s' % (flavour, description)
        self.addLine(complaint)
#       inspectObject(test)
        thing = getattr(test, test._TestCase__testMethodName)
        file, line = _createSourceTuple(thing)
        testAt = '%s:%i: %s' % (file, line, test._TestCase__testMethodName)
        self.addLine(testAt)

    #  unitTestFile helps censor the redundant line
    #  in the traceback where unittest.py threw us the exception...

        unitTestFile = iz.getsourcefile(TestSuite)
        unitTestFile = '  File "%s", line' % unitTestFile

        for line in apply(traceback.format_exception, err):
            if line[:len(unitTestFile)] != unitTestFile:
                for l in string.split(line,"\n")[:-1]:
                    self.addLine(l)
        return

    def write(self, lines):
        self.wazStderr.write(lines)

        for line in string.split(lines, '\n'):
            self.addLine(line)

    def addLine(self, line):
        self.myErrorList.insert('end', line)

    def getDescription(self, test):
        return test.shortDescription() or str(test)

    def addError(self, test, err):
        _TextTestResult.addFailure(self, test, err)
        self._printErrorList('ERROR', test, err)
        self.addEvent()

    def getFraction(self):
        fraction = self.testsRun / float(self.myTotal)
        return fraction

    def getColor(self):
        color = 'cyan'

        if len(self.errors) > 0:
            color = 'red'
        elif len(self.failures) > 0:
            color = 'orange'
        elif self.testsRun / self.myTotal:
            color = 'green'
            sys.stderr = self.wazStderr  #  sick but ... tested ;-)

        return color


class BrowserNode:
    'Behavior & interface common to any clickable in the tree'

    def __init__(self, browser, name):
        self.myChildren = []
        self.myChildrensNames = {}
        self.stream = _WritelnDecorator(sys.stdout)
        self.myName = name
        self.myBrowser = browser

    def getName(self):  
        assert self.myName != ''
        return self.myName

    def getChild(self, idx):
        if idx >= len(self.myChildren):  return None
        return self.myChildren[idx]

    def addChild(self, node):
        'This is a job for an "ordered map"'
        
        name = node.getName()
        assert name is not None
        assert name != ''
        got = self.myChildrensNames.get(name, None)
        if got is not None:  return got
        self.myChildren.append(node)
        self.myChildrensNames[name] = node
        return node

    def getDoc(self):  return iz.getdoc(self.getThing())
    def getSourceLocation(self, extra = 0):  return (None, None)
    def uniqueness(self):  return ''
    def getTestCount(self):  return self.getThing().countTestCases()
    def augmentMenu(self, rmenu):  pass
    def pyCheck(self):  pass
    
    def growTree(self, aTree, parentKey = ''):
        name = self.getName()
        key = parentKey + _slash + name + self.uniqueness()
        aTree.AddNode(key, name, parentKey, Data = self)
        
        for child in self.myChildren:
            child.growTree(aTree, key)
        
        return

    def getThing(self):
        ''' If we are not a TestSuite or TestCase, build a TestSuite
            to represent what we are '''
            
        aSuite = TestSuite()
        
        for suite in self.myChildren:
            aSuite.addTest(suite.getThing())
            
        return aSuite

    def _makeResult(self):  return BrowserResult(self.stream)
    def checkTime(self):  return 0  #  non-file things never change their time

    def checkTimes(self):
        anyChanged = self.checkTime()
        
        for child in self.myChildren:
            if child.checkTimes():
                anyChanged = 1
                
        return anyChanged

    def grep(self, slug):
        files = self.getAllFileNames()
        command = ['grep -Hn'] + [slug] + files
        command = string.join(command, ' ')
        print command
        self.myBrowser.myErrorList.clear()
        self.myBrowser.myBar.setMessage(command)
        grepErrorCode = piper.Piper(command, self.outputSink, self.outputSink)
        return grepErrorCode
          
    def getAllFileNames(self):
        file, line = self.getSourceLocation()
        source, line = self.getSourceLocation(1)
        if source is None or source == file:  return [file]
        return [file, source]
        
    def outputSink(self, lines):
        'TODO have we seen this before?'
        aList = self.myBrowser.myErrorList
        
        for line in string.split(lines, '\n'):
            line = string.strip(line)
            if len(line) > 0:
                aList.insert('end', line)
                aList.update()
        return
           
    def callTests(self, result):
        startTime = time.time()
        self.getThing()(result)
        stopTime = time.time()
        timeTaken = float(stopTime - startTime)
        result.printErrors()
        self.stream.writeln(result.separator2)
        run = result.testsRun

        self.stream.writeln("Ran %d test%s in %.3fs" %
                            (run, run == 1 and "" or "s", timeTaken))

        self.stream.writeln()

        if not result.wasSuccessful():
            self.stream.write("FAILED (")
            failed, errored = map(len, (result.failures, result.errors))
            
            if failed:
                self.stream.write("failures=%d" % failed)
                
            if errored:
                if failed: self.stream.write(", ")
                self.stream.write("errors=%d" % errored)

            self.stream.writeln(")")
        else:
            self.stream.writeln("Okay")

        return


def _createSourceTuple(thing):
    'this wastes all of [0] - screw it!'
    return (iz.getsourcefile(thing), iz.findsource(thing)[1])


def _findFunctionSource(functions, methodName):
    for name, funk in functions:
        if name == methodName:
            return _createSourceTuple(funk)


class TreeTestCase(BrowserNode):

    def __init__(self, browser, aCase):
        
#  should our getThing return getMethod?
        
        BrowserNode.__init__(self, browser, '')
        self.myCase = aCase

    def getDoc(self):
        return self.myCase._TestCase__testMethodDoc
   
    def getLineCount(self):
        #print self.getMethod()
        loc = iz.findsource(self.getMethod())
        lines = string.split(loc[0], '\n')
        print lines[loc[0]]
 
        return 4
    
    def getExtraLocation(self, extra = 0):
        sourceFile = iz.getsourcefile(self.getThing().__class__)
        module = getTesteeModule(sourceFile)

        if module:
    
    #  observe we assume all test methods are named after their
    #  testee method prefixed with test_

            methodName = self.getName()[len('test_'):]
            functions = iz.getmembers(module, iz.isfunction)
            source = _findFunctionSource(functions, methodName)
            if source is not None:  return source
            classes = iz.getmembers(module, iz.isclass)

            for className, klass in classes:
                if className == methodName:
                    # you can name a test method after a class

                    source = _createSourceTuple(klass)
                    if source is not None:  return source

                functions = iz.getmembers(klass, iz.ismethod)
                source = _findFunctionSource(functions, methodName)
                if source is not None:  return source

        return (None, None)

    def getSourceLocation(self, extra = 0):
        if extra == 1:
            return self.getExtraLocation()

        method = self.getMethod()
        return _createSourceTuple(method)

    def getMethod(self):
        return getattr(self.getThing(), self.getName())

    def getName(self):  return self.myCase._TestCase__testMethodName
    def getThing(self):  return self.myCase
    def getTestCount(self):  return 1


def _traverseTestList(aBrowser, tests):
    children = []

    for test in tests:
        if test.__class__.__name__ == 'TestSuite':
            aSuite = TreeTestSuite(aBrowser, test)
            q = 0

            for child in children:
                if child.getName() == aSuite.getName():
                    q += 1
    
            if q > 0:  aSuite.myUniqueness = str(q)
            children.append(aSuite)
        else:
            aCase = TreeTestCase(aBrowser, test)
            children.append(aCase)
    
    return children

    
class TreeTestSuite(BrowserNode):
    def __init__(self, browser, aSuite):
        BrowserNode.__init__(self, browser, '')
        self.mySuite = aSuite
        assert aSuite.__class__.__name__ == 'TestSuite'
        self.myChildren += _traverseTestList(self.myBrowser, self.mySuite._tests)
        self.myUniqueness = ''
        
    def getSourceLocation(self, extra = 0):
        if extra == 1:  return (None, None)
        return _createSourceTuple(self.getThing().__class__)

    def getAllFileNames(self):
        if self.myChildren == []:  return []
        thing = self.myChildren[0].getMethod()
        file = iz.getsourcefile(thing)
        return [file]

    def getName(self):  return self.mySuite.__class__.__name__
    def getThing(self):  return self.mySuite
    def uniqueness(self):  return self.myUniqueness
    def getTestCount(self):  return self.mySuite.countTestCases()


class TestFile(BrowserNode):

    def __init__(self, filePath, browser):
        BrowserNode.__init__(self, browser, '')
        self.myFilePath = filePath
        self.myModule = _reImport(self.myFilePath)
        self.myModuleTime = path.getmtime(self.myFilePath)
        self.mySourceTime = self.getSourceTime()
        self._findMySuite()

    def _findMySuite(self):
        funcs = iz.getmembers(self.myModule, iz.isfunction)
        myTestSuite = None

        for name, method in funcs:
            if name == 'suite':
                myTestSuite = method()
                self.myChildren += _traverseTestList(self.myBrowser, myTestSuite._tests)

        if myTestSuite is None:
            myTestSuite = findTestCases(self.myModule)

            if myTestSuite._tests:
                self.myChildren += _traverseTestList(self.myBrowser, myTestSuite._tests)
            else:
                if aCommandLine.getVerbose():
                    print 'rejecting', self.getName()

                self.myModule = None  #  suicide

        return

    def getSourceTime(self):
        file, line = self.getSourceLocation(extra = 1)
        if file is None:  return None
        return path.getmtime(file)

    def checkTime(self):
        nuTime = path.getmtime(self.myFilePath)
        nuSourceTime = self.getSourceTime()

        if self.myModuleTime == nuTime and self.mySourceTime == nuSourceTime:
            return 0

        self.myModuleTime = nuTime
        self.mySourceTime = nuSourceTime
        time.sleep(0.1)  #  time for write to finish!
        self.myModule = _reImport(self.myFilePath)
        self.myChildren = []
        self._findMySuite()
        file, line = self.getSourceLocation(extra = 1)
        if file is not None:  _reImport(file)  #  TODO use a member!
        return 1

    def augmentMenu(self, rmenu):
        rmenu.add_command(
                label = "PyChecker %s - <F7>" % (self.getName()),
                command = self.pyCheck
                )
        file, line = self.getSourceLocation(extra = 1)
        if file is None:  return

        rmenu.add_command(
                label = "PyChecker %s - <F8>" % (file),
                command = self.pyCheckSource
                )
        return
        
    def pyCheck(self):  
        return self.pyCheckFile(self.getName())
    
    def pyCheckSource(self):    
        fileName, line = self.getSourceLocation(extra = 1)
        if fileName is None:  return
        return self.pyCheckFile(fileName)
        
    def pyCheckFile(self, fileName):
        self.myBrowser.myErrorList.clear()
        
        command = 'pychecker ' + fileName
        self.myBrowser.myBar.setMessage(command)
        got = piper.Piper(command, self.outputSink, self.errorSink)
        self.myBrowser.myBar.setMessage(command)

        if got == 0:
            self.myBrowser.myBar.setMessage('No warnings', 'green')
        else:
            self.myBrowser.myBar.setMessage('Warnings', 'orange')

        return got

    def outputSink(self, lines):
        for line in string.split(lines, '\n'):
            line = string.strip(line)

            if len(line) > 0:
                self.myBrowser.myErrorList.insert('end', line)
                self.myBrowser.myErrorList.update()
        return

    def errorSink(self, lines):
        #self.outputSink(lines)
        for line in string.split(lines, '\n'):
            sys.stderr.write(line)

    def getName(self):
        return path.split(iz.getsourcefile(self.myModule))[1]

    def getDoc(self):  return iz.getdoc(self.myModule)

    def getSourceLocation(self, extra = 0):
        if extra:
            module = getTesteeModule(self.myFilePath)
            
            if module:
                return (iz.getsourcefile(module), 0)
                
            return (None, None)
            
        return (self.myFilePath, 0)


def getTesteeModule(fileName):
    'if your naming conventions differ, edit here'

    # TODO  make sure this crap does not target any "test" in the folder names

#  TODO look in the testcase module itself too
    
#   assert string.count(fileName, 'test') > 0

    if string.count(fileName, 'test_') > 0:
        testeeFile = string.join(string.split(fileName, 'test_'), '')
    elif string.count(fileName, '_test') > 0:
        testeeFile = string.join(string.split(fileName, '_test_'), '')
    else:
        testeeFile = string.join(string.split(fileName, 'test'), '')
                
    if not path.exists(testeeFile):  return None
    module = _reImport(testeeFile)
    return module


def _reImport(filePath):
    'TODO why we call this twice per module?'

    if filePath == 'browser.py':
        import browser
        return browser

    folder, mod = path.split(filePath)

  #  TODO  use some path.snipextension here

    mod = mod[:-3]
    got = imp.find_module(mod, [folder])

    module = apply(imp.load_module, [mod] + list(got))
    return module


_syntaxErrorLine = re.compile(r'^  File "(.*)", line ([0-9]*)')
_testFailureLine = re.compile(r'^(.*):([0-9]*):')


def identifyErrorLine(line):
    'finds file info in a standard error line'

    got = _syntaxErrorLine.search(line)

    if got is None:
        got = _testFailureLine.search(line)

        if got is None:
            return None, None

    group = got.groups(0)

    if len(group) != 2:
        return None, None

    file = group[0]
    line = int(group[1])

    return file, line


class FolderNode(BrowserNode):

    def __init__(self, folder, browser):
        BrowserNode.__init__(self, browser, folder)
        self.myInitModule = None
   
    def getDoc(self):
        if self.myInitModule is None:
            initPath = path.join(self.myName, '__init__.py')
            
            if not path.exists(initPath):
                return 'no __init__.py file to read its doc string'
          
            self.myInitModule = _reImport(initPath)  
            
        self.myInitModule = reload(self.myInitModule)  #  TODO do we reload enough?
        return iz.getdoc(self.myInitModule)
   
    def getAllFileNames(self):
        names = []
   
        for thing in self.myChildren:
            names += thing.getAllFileNames()
            
        return names
    
    def findAppropriateFolder(self, folder):
        aThing = TestFolderName(folder, self.myBrowser)
        aThing = self.addChild(aThing)
        return aThing


class TestFolder(FolderNode):
    ''' This, the dot . , is so you can run all modules in a 
        folder but not the sub folders '''
    
    def __init__(self, browser):
        FolderNode.__init__(self, _dot, browser)
   
    def absorbFile(self, fileName):
        aFile = TestFile(fileName, self.myBrowser)
        if aFile.myModule is None:  return
        self.addChild(aFile)
                

class TestFolderName(FolderNode):
    'This is so you can run a folder and all sub-folders in it'
    
    def __init__(self, folder, browser):
        FolderNode.__init__(self, folder, browser)
               
    def absorbFile(self, parts, fileName):
        if len(parts) > 1:
            aThing = self.findAppropriateFolder(parts[0])
            aThing.absorbFile(parts[1:], fileName)
            return

        aFolder = TestFolder(self.myBrowser)
        aFolder = self.addChild(aFolder)
        aFolder.absorbFile(fileName)


class All(TestFolderName):
    'the <all> object'
    
    def __init__(self, files, browser):
        TestFolderName.__init__(self, '<all>', browser)
        map(self.absorbAllFiles, files)
        
    def absorbAllFiles(self, fileName):
        assert fileName[0] != '.'
        parts = string.split(fileName, os.sep)
        self.absorbFile(parts, fileName)      

    def getDoc(self):  return 'Tap <Enter> to run all tests'
    

class Browser:
    'The owner of the UI elements'
    
    def __init__(self, files):
        'TODO: check for the "nothing to test!" condition'
        
        self.myAll        = All(files, self)
        self.myBar        = None
        self.myTree       = None
        self.myErrorList  = None
        self.myFrame      = None
        self.myEditorMap  = {}

    def initializeTree(self):
        '''put the keyboard focus on the tree (you'd be surprised
           how many UIs forget this detail), then expand it 2 ranks.'''
                
        self.myTree.focus_set()
        self.expandRoot()
        self.timer()
        
    def expandRoot(self):
        key1 = _slash + self.myAll.getName()
        key2 = key1 + _slash + _dot
        self.myTree.Expand(key1)
        self.myTree.Expand(key2)
        self.myTree.Select(key1)

    def reStuffTree(self):      
        '''always reselects >something<, but will only 
           return 1 if the current node did not go away'''
        
        aNode = self.myTree._SelectedNode
        key = aNode['Key']
        formerKey = key
        expanded = aNode['Expanded']
        self.myTree.Clear()
        self.myAll.growTree(self.myTree)
        self.expandRoot()

        while len(key) > len(self.myAll.getName()):
            if self.myTree.Select(key):
                if expanded or key != formerKey:
                    self.myTree.Expand(key)

                return key == formerKey

            key = string.join(string.split(key, _slash)[:-1], _slash)
        
        return 0  #  current node went away
                
    def timer(self):
        if self.myAll.checkTimes():
            if self.reStuffTree():
                aNode = self.myTree.GetSelectedNode('Data')
                self.runSelectedTests(aNode)
                self.startTimer()
        else:

    #  observe this only restarts the timer if the
    #  reloads succeeded

            self.startTimer()

    def startTimer(self):
        self.myTree.after(500, self.timer)

    def browse(self, root):
        root.title(_getCurrentFolder() + " Test Browser")
        self.myFrame = root
        panes = Pmw.PanedWidget(root)
        top = panes.add('top')
        bottom = panes.add('bottom')
        aTreeBox = self._buildTree(top)
        self.myTree = aTreeBox
        _packMax(aTreeBox)
        self.myAll.growTree(aTreeBox)

        self.myErrorList = Pmw.ScrolledListBox(
                bottom,
                 dblclickcommand = self.editErrorInList,
                 #     command = self.editErrorInList,
                selectioncommand = self.editErrorInList,
                )
        self.myBar = ProgressBar(bottom)
        self.myBar.pack(fill = 'x')
        _packMax(self.myErrorList)
        _packMax(panes)
        
        self.myGrepField = Pmw.EntryField(
                                   root, 
                                   labelpos = 'w',
                                   label_text = 'grep:',
                                   entry_width = 40,
                                   command = self.grepFiles,
                                   )

    def editErrorInList(self, args = None):
        args = args
        sel = self.myErrorList.getcurselection()
        if len(sel) != 1:  return
        errorLine = sel[0]
        file, line = identifyErrorLine(errorLine)
        if file is None or line is None:  return
        file = _completePath(file)
        self.editSource(file, line)

    def treeNodeClick(self, arg):
        node = self._treeData(arg)
        doc = node.getDoc()
        self.myBar.setMessage(doc, 'white')
    
    def runSelectedItem(self, arg):  
        node = self._treeData(arg)    
        if node is None:  return
        self.runSelectedTests(node)
        
   #  The timer only starts at launch, and after you <Enter>
   #  or double-click a node, and if the timer has successfully run.
   #  If the timer experienced an error reloading modules, it stays
   #  off until the next time you manually test (or restart, of course)     
        
        self.startTimer()
        
    def runSelectedTests(self, node):

#  TODO rename to myOutputList
        
        self.myErrorList.clear()
        result = node._makeResult()
        testCount = node.getTestCount()
        result.setUserView(self.myBar, self.myErrorList, testCount)
        node.callTests(result)

    def addEditCommand(self, rmenu, file, line, concept, accel):
        "Houston, we've had a Lambda"
        
        if file is None or line is None:  return
        file = _completePath(file)
        line += 1  #  some editors are 1-based

        def command(self = self, file = file, line = line):
            self.editSource(file, line)

        rmenu.add_command(
                label = "Edit %s `%s' at %i - <F%i>" % (concept, file, line, accel),
                command = command
                )
                
    def treeContextMenu(self, arg):
        node = self._treeData(arg)
        thing = node.getThing()
        rmenu = tk.Menu(self.myFrame, tearoff = 0)
        file, line = node.getSourceLocation()
        
        #  following "self." may fix Richard Jones's insect
        
        self.addEditCommand(rmenu, file, line, 'test', 2)
        file, line = node.getSourceLocation(1)
        self.addEditCommand(rmenu, file, line, 'code', 3)
        node.augmentMenu(rmenu)        
        rmenu.add_command(label = "grep - <F9>", command = self.showGrepField)
        rmenu.tk_popup(arg.x_root, arg.y_root)

    def showGrepField(self, event):
        self.myGrepField.pack()
        self.myGrepField.component('entry').focus_set()

    def grepFiles(self, event = None):
        aNode = self.myTree._SelectedNode
        node = aNode['Data']
        node.grep(self.myGrepField.get())

    def editSourceAccelKey(self, event):
        node = self._treeData(event)
        
        if event.keysym == 'F7':
            node.pyCheck()
            return
                       
        if event.keysym == 'F8':
            node.pyCheckSource()
            return
            
        sourceType = event.keysym == 'F3'
        if node is None:  return
        file, line = node.getSourceLocation(sourceType)
        if file is None or line is None:  return
        self.editSource(file, line)

    def editSource(self, file, line, bp = 0):
        use = aCommandLine.getEditor()
        
        if use == USE_VIM and not bp:
            slug = 'vim -g --servername TDD +%i %s &'
                    
            cmd = slug % (line, file)
            print cmd
            os.system(cmd)
        elif use == USE_KATE:
            slug = "kate --line %i %s &"
                    
            cmd = slug % (line, file)
            print cmd
            os.system(cmd)           
        else:
            from idle.PyShell import PyShellEditorWindow
            from idle.PyShell import PyShellFileList
            top = Pmw.initialise()
            
            aEditor = self.myEditorMap.get(file, None)

            #  TODO  let the user close the danged window
            #  TODO  remap all its keystrokes to something humane
            #  TODO  detect when the window got closed and remove it from the map
            #  TODO  plug into the debugger!
            
            if aEditor is None:
                # >sigh< why do I have to give you a steeking FileList?
                aFileList = PyShellFileList(top)
                aEditor = PyShellEditorWindow(aFileList, file, None, top)
                self.myEditorMap[file] = aEditor

#           inspectObject(aEditor)
            aEditor.gotoline(line + 1)
            aEditor.top.tkraise()
            aEditor.top.focus_force()
            aEditor.text.focus_force()

            # what part of "raise" don't you understand??
            
            if bp:
                aEditor.set_breakpoint_here()
            
#           inspectObject(aEditor.top)

        return
        
    def _treeData(self, arg):
        tree = arg.widget
        if tree._SelectedNode is None:  return None
        return tree.GetSelectedNode('Data')
               
    def _buildTree(self, parent):
        tree = TreeBox(
                        parent,
                        visibleicons = tk.NO, # YES,
                        bg = '#eeeeee',
                        linecolor = 'black',
                        selcolor = 'lightgreen',
                        signcolor = 'red',
                        font = 'helvetica 14 bold',
                        )
        
        tree.focus_set()
        tree.bind('<<NodeClick-1>>', self.treeNodeClick)
        tree.bind('<<Select>>', self.treeNodeClick)
        tree.bind('<<Return>>', self.runSelectedItem)
        tree.bind('<<NodeClick-2>>', self.treeContextMenu)
        tree.bind('<<NodeClick-3>>', self.treeContextMenu)
        tree.bind('<F2>', self.editSourceAccelKey)
        tree.bind('<F3>', self.editSourceAccelKey)
        tree.bind('<F7>', self.editSourceAccelKey)
        tree.bind('<F8>', self.editSourceAccelKey)
        tree.bind('<F9>', self.showGrepField)
        return tree


def _packMax(widget):
    widget.pack(expand = 1, fill = 'both')


def _completePath(file):
    return path.normcase(path.abspath(file))


class ProgressBar(tk.Frame):
    """A simple progress bar that shows a percentage progress in
    the given colour."""

    def __init__(self, *args, **kwargs):
        apply(tk.Frame.__init__, (self,) + args, kwargs)

        self.myCanvas = tk.Canvas(
            self,
            height = 32,
#          width = _width,
            background = 'white',
            borderwidth = 3,
            )
        self.myCanvas.pack(fill=tk.X, expand=1)
        self.rect = self.text = None
        # self.myCanvas.bind('<Configure>', self.paint)
        self.setProgressFraction(0.0, color = 'white')
        self.text = None

    def setProgressFraction(self, fraction, color):
        self.fraction = fraction
        self.color = color
        self.paint()
        self.myCanvas.update_idletasks()

    def paint(self, *args):
        percentString = "%3.0f%%" % (100.0 * self.fraction)
        self.setMessage(percentString)

    def setMessage(self, message, color = None):
        if color is not None:
            self.color = color

        cnv = self.myCanvas
        totalWidth = cnv.winfo_width()
        width = int(self.fraction * float(totalWidth))
        height = cnv.winfo_height()
        if self.rect is not None:  cnv.delete(self.rect)
        if self.text is not None:   cnv.delete(self.text)

        self.rect = cnv.create_rectangle(0, 0, width, height,
                                                 fill = self.color)

        self.text = cnv.create_text(
                        totalWidth / 2,
                        height / 2,
                        anchor = tk.CENTER,
                        text = message
                )
        self.myCanvas.update_idletasks()
        return


def _getCurrentFolder():
    dirname = os.getcwd()
    ignoreMe, dirname = path.split(dirname)
    return dirname


if __name__ == '__main__':
    aCommandLine.absorbCommandLine(relevantArgs)
    root     = Pmw.initialise()
    files    = aCommandLine.getFiles()
    aBrowser = Browser(files)
    aBrowser.browse(root)
    root.after_idle(aBrowser.initializeTree)
    tk.mainloop()

