#!/usr/bin/env python
#
#  LGPL - don't steal, don't sue
#
'tests the Browser test runner'

import unittest, os, inspect, time, string, sys
from os import path
from browser import Browser
from browser import identifyErrorLine
from browser import CommandLineParser, USE_KATE, USE_IDLE, USE_VIM
from mock import Mock
from Debug import inspectObject, tolerance, db, removeForcedRecursive

iz = inspect


class BrowserTestCase(unittest.TestCase):
    """Tests the TestBrowser class"""

    def setUp(self):
        self.tearDown()

    def test_getChild(self):
        aBrowser = _makeCensoredBrowser()
        self.assertEquals(aBrowser.myAll.getName(), '<all>')
        aChild = aBrowser.myAll.getChild(0)
        
    #  TODO observe things are generally in alpha order      

        expecting = [   '.', 
                        'alltests.py',
                        'TestSuite',
                        'TestSuite',
                        'testAppend', ]
        
        for expect in expecting:
            self.assertEquals(aChild.getName(), expect)
            aChild = aChild.getChild(0)

        self.assertEquals(aChild, None)           
        aChild = aBrowser.myAll.getChild(1)
        self.assertEquals(aChild, None)

    def test_growTree(self):
        aBrowser = _makeCensoredBrowser()
        aTreeBox = Mock()
        aBrowser.myAll.growTree(aTreeBox)
        calls = aTreeBox.getAllCalls()

        for call in calls:
            self.assertEquals(call.getName(), 'AddNode')
            self.assertEquals(len(call.params), 3)
            self.assertEquals(len(call.kwparams), 1)
            snip = repr(call.kwparams['Data'])[:9]
            self.assertEquals(snip, '<browser.')
            
        x  = 0;  self.assertEquals(calls[x].params, ('/<all>', '<all>', ''))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/.', '.', '/<all>'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py', 'alltests.py', '/<all>/.'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite', 'TestSuite', '/<all>/./alltests.py'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite', 'TestSuite', '/<all>/./alltests.py/TestSuite'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite/testAppend', 'testAppend', '/<all>/./alltests.py/TestSuite/TestSuite'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite/testCount', 'testCount', '/<all>/./alltests.py/TestSuite/TestSuite'))

    def test_callTests(self):
        aBrowser = _makeCensoredBrowser()
        aTestCase = aBrowser.myAll.getChild(0).getChild(0).getChild(0).getChild(0).getChild(0)
        self.assertEquals(aTestCase.getName(), 'testAppend')
        self.assertEquals(aTestCase.getTestCount(), 1)
        result = aTestCase._makeResult()
        aTestCase.callTests(result)
        self.assertEquals(result.testsRun, 1)

        aTestSuite = aBrowser.myAll.getChild(0).getChild(0).getChild(0).getChild(0)
        self.assertEquals(aTestSuite.getName(), 'TestSuite')
        self.assertEquals(aTestSuite.getTestCount(), 4)
        result = aTestSuite._makeResult()
        aTestSuite.callTests(result)
        self.assertEquals(result.testsRun, 4)

        aFolder = aBrowser.myAll.getChild(0)
        self.assertEquals(aFolder.__class__.__name__, 'TestFolder')
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getName(), '.')

        self.assertEquals(aFolder.getTestCount(), 416)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 416)
        aFile = aBrowser.myAll.getChild(0).getChild(0)
        self.assertEquals(aFile.__class__.__name__, 'TestFile')
        self.assertEquals(aFile.getTestCount(), 8)
        self.assertEquals(aBrowser.myAll.getTestCount(), 416)

    def test_GreenBar(self):
        aBrowser = _makeCensoredBrowser()
        aFolder = aBrowser.myAll.getChild(0)      
        bar, lister = _mockProgress(aFolder)
        calls = bar.getAllCalls()
        
        for call in calls:
            self.assertEquals(call.getName(), 'setProgressFraction')

        for call in calls[:-1]:
            self.assertEquals(len(call.params), 2)
            fraction = call.getParam(0)
            self.assert_(fraction >= 0.0 and fraction <= 100.0)
            color = call.getParam(1)
            self.assertEquals(color, 'cyan')

        self._testEndOfBar(calls, 'green')
        self.assertEquals(len(calls), 416)

    def _testEndOfBar(self, calls, endColor):
        call = calls[-1]
        self.assertEquals(len(call.params), 2)
        fraction = call.getParam(0)
        self.assert_(fraction >= 0.9999 and fraction <= 1.001)
        self.assert_(tolerance(fraction, 1.0))
        color = call.getParam(1)
        self.assertEquals(color, endColor)

    def test_OrangeBar(self, problem = 'self.assert_(0)', color = 'orange'):
        open(_test_orangeHerring, 'w') \
            .write(_instaTest % problem)
            
        aBrowser = Browser(getFiles(['.']))
        aFile = aBrowser.myAll.getChild(0).getChild(5)
        fileName = aFile.getName()
        compare = _test_orangeHerring
        
        if fileName[0:2] == './':
            compare = './' + compare

        self.assertEquals(fileName, compare)
        bar, lister = _mockProgress(aFile)
        calls = bar.getAllCalls()
        self._testEndOfBar(calls, color)
        self.assertEquals(len(calls), 1)
        self.tearDown()
        calls = lister.getAllCalls()
        self.assertNotEquals(calls, [])

        for call in calls:
            self.assertEquals(call.getName(), 'insert')

        self.assertEquals(calls[0].params, ('end', 'FAIL: testFaylur (test_orangeHerring.RedHerringTestCase)'))
        self.assertEquals(calls[2].params, ('end', 'Traceback (most recent call last):'))

        either = ('end', '  File "test_orangeHerring.py", line 5, in testFaylur')
        ore = ('end', '  File "./test_orangeHerring.py", line 5, in testFaylur')

        self.assert_(calls[3].params == either or calls[3].params == ore)

        param4 = calls[4].params
        self.assertEquals(param4[0], 'end')
        self.assertEquals(string.strip(param4[1]), 'self.assert_(0)')

    def teest_RedBar(self):
        'TODO does anyone know how to create an error?'
        self.test_OrangeBar('x = 1 / 0', 'red')

    def test_subFolders(self):
        _putPassingTestInSubFolder(withInit = 0)
        aBrowser = Browser(getFiles(['.']))
        aBrowser.myAll.getChild(1)
        
    # TODO we decline to observe files without a __init__.py

#        self.assertEquals(aFolder, None)
        self.tearDown()
        _putPassingTestInSubFolder(withInit = 1)
        
        aBrowser = Browser(getFiles(['.']))
        aFolderName = aBrowser.myAll.getChild(0)
        ''' TODO
        self.assertEquals(aFolderName.getName(), _subFolder)
        aFolder = aFolderName.getChild(0)
        aFile = aFolder.getChild(0)
        self.assertEquals(aFile.getName(), _testBullfrog)
        '''

    #  uncomment this to browse the scratch folders
#        os.system('python ./browser.py')

    def test_setUp(self):
        'run a test that only passes if setUp works'

        instaTest = '''
import unittest, sys
x = 0
class RedHerringTestCase(unittest.TestCase):
    def setUp(self):
        global x
        x = 1
    def testFaylur(self):
        self.assertEquals(x, 1)
''' + _instaSuite

        os.mkdir(_subFolder)
        open(_subFile, 'w').write(instaTest)

    # This tests that Browser obeys folders other than '.'

        aBrowser = Browser(getFiles([_subFolder]))
        aFolder = aBrowser.myAll.getChild(0)
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getName(), _subFolder)
        self.assertEquals(aFolder.getChild(0).getName(), '.')
        self.assertEquals(aFolder.getChild(0).getChild(0).getName(), _testBullfrog)
        self.assertEquals(aFolder.getTestCount(), 1)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 1)

    def test_tearDown(self):
        'run a tearDown and inspect it for side-effects'

        instaTest = '''
import unittest, sys
x = 0
class RedHerringTestCase(unittest.TestCase):
    def testFaylur(self):
        pass
    def tearDown(self):
        self.instaMember = 1
''' + _instaSuite

        os.mkdir(_subFolder)
        open(_subFile, 'w').write(instaTest)

    # This tests that Browser obeys folders other than '.'
        
        aBrowser = Browser(getFiles([_subFolder]))
        aFolder = aBrowser.myAll.getChild(0)
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getTestCount(), 1)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 1)
        aRedHerring = aFolder.getChild(0).getChild(0).getChild(0).getChild(0).getThing()
        self.assertEquals(aRedHerring.instaMember, 1)

    # was that sick or what? ;-)

    def test_getDoc(self):  #  also probes getSourceLocation
        'my doc'
        
        try:
            raise Exception('yo')
        except:
            f, file, lineno, fn_name, lines, n = iz.trace()[-1]
            lineno -= 7
            
    #   lineno = 246  #  yes if you edit above here you must change this

        aBrowser = Browser(getFiles(['.']))
        idx = 0
        thisTest = None
        aFolder = aBrowser.myAll.getChild(0)

    #  a folder's doc string is that doc from its __init__.py file, if any

        self.assertEquals(aFolder.getDoc(), 'illogical')
        self.assertEquals(aFolder.getSourceLocation(), (None, None))

    #  note nobody needs iterators except this test, so far

        while aFolder.getChild(idx) is not None and thisTest is None:
            if aFolder.getChild(idx).getName() == 'test_browser.py':
                aFile = aFolder.getChild(idx)
                self.assertEquals(aFile.getDoc(), 'tests the Browser test runner')
                self.assertEquals(aFile.getSourceLocation(), (aFile.myFilePath, 0))
                aSuite = aFile.getChild(0)
                file, line = aSuite.getSourceLocation()
                ignore, fileName = path.split(file)
                self.assertEquals(fileName, 'unittest.py')
                self.assert_(line > 100 and line < 1000)
                self.assertEquals(aSuite.getSourceLocation(1), (None, None))
                idx2 = 0

                while aSuite.getChild(idx2) is not None and thisTest is None:
                    aTest = aSuite.getChild(idx2)

                    if aTest.getName() == 'test_getDoc':
                        file, line = aTest.getSourceLocation()
                        expect = 'test_browser.py'
                        self.assertEquals(file[-len(expect):], expect)
                        self.assertEquals(line, lineno)
                        thisTest = aTest
            
                    idx2 += 1

            idx += 1

        self.assertNotEquals(thisTest, None)
        doc = thisTest.getDoc()
        self.assertEquals(doc, 'my doc')
                
    def test_checkTime(self):
        open(_test_orangeHerring, 'w') \
            .write(_instaTest % 'pass')

        aBrowser = Browser(getFiles(['.']))
        aFile = aBrowser.myAll.getChild(0).getChild(5)
        wazModule = aFile.myModule
        got = aFile.checkTime()
        self.assert_(not got)
        self.assertEquals(wazModule, aFile.myModule)

    #  now add a test, and change the time

        open(_test_orangeHerring, 'w') \
            .write(_instaTest % 'pass\n    def test_zone(self):\n        pass')

        file, line = aFile.getSourceLocation()
        self.assertNotEquals(file, None)
        same = path.samefile(file, _test_orangeHerring)

        if not same:
            db(file)
            db(_test_orangeHerring)
            
        self.assert_(same)
        
        touch(_test_orangeHerring, time.time() - 1000000)
        got = aFile.checkTime()
        self.assert_(got)
        
#  We can't check this because 'imp.loadModule' is so aggressive
#  it replaces the guts of the module object /in-situ/.
#        self.assertNotEquals(wazModule, aFile.myModule)

        self.assertEquals(iz.getfile(wazModule), iz.getfile(aFile.myModule))
        aSuite = aFile.getChild(0)
        aTest = aSuite.getChild(0)
        self.assertNotEquals(aTest, None)
        bTest = aSuite.getChild(1)       
        self.assertNotEquals(bTest, None)
        self.assert_('test_zone' in [aTest.getName(), bTest.getName()])
        
    #  now peek at the time again, and show checkTime has
    #  a side-effect

        got = aFile.checkTime()
        self.assert_(not got)

    def test_identifyErrorLine(self):
        example = '''Traceback (most recent call last):
  File "./testrunner.py", line 16, in ?
    got = runMe()
  File "./OTestRunner.py", line 1196, in runMe
    return t.run()
  File "./OTestRunner.py", line 1128, in run
    self.runGui()
  File "./OTestRunner.py", line 919, in runGui
    suiteList.stuffTree()
  File "./OTestRunner.py", line 734, in stuffTree
    aFolder = TestFolder('.')
  File "./OTestRunner.py", line 220, in __init__
    self.collectFiles()
  File "./OTestRunner.py", line 250, in collectFiles
    aModule = TestModule(f)
  File "./OTestRunner.py", line 336, in __init__
    self.loadModule()
  File "./OTestRunner.py", line 377, in loadModule
    aClass = TestClass(aKey, klass, self)
  File "./OTestRunner.py", line 461, in __init__
    aMethod = TestMethod(self, aKey)
  File "./OTestRunner.py", line 523, in __init__'''

        lines = string.split(example, '\n')

        lineNumbers = [
            16, 1196, 1128, 919, 734,
            220, 250, 336, 377, 461, 523
            ]

        for idx in range(0, len(lines), 2):
            file, line = identifyErrorLine(lines[idx])
            self.assertEquals(file, None)
            self.assertEquals(line, None)
            file, line = identifyErrorLine(lines[idx+1])

            if idx == 0:
                self.assertEquals(file, './testrunner.py')
            else:
                self.assertEquals(file, './OTestRunner.py')

            self.assertEquals(line, lineNumbers[idx / 2])

        sample = './OTestRunner_.py:721: TEST(0) Failed.'
        file, line = identifyErrorLine(sample)
        self.assertEquals(file, './OTestRunner_.py')
        self.assertEquals(line, 721)

    '''def teest_editSource(self):
        #  TODO  introduce rabid 'if' statements here to recuse
        #       the test if idle is not installed
        aBrowser = Browser(['.'])
        #aBrowser.editSource('browser.py', 42 )'''

    def teest_getLineCount(self):
        aBrowser = _makeCensoredBrowser()
        aSuite = aBrowser.myAll.getChild(0)
        aSuite = aSuite.getChild(0)
        aSuite = aSuite.getChild(0)
        aSuite = aSuite.getChild(0)
        aTest = aSuite.getChild(0)
        self.assertEquals(aTest.getName(), 'testAppend')
        self.assertEquals(aTest.getLineCount(), 3)

    def test_CommandLineParser(self):
        sys_args = ['ignoreMe',  '--vim', 'test*.py', '..', 'test_browser.py']
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(sys_args[1:])
        self.assertEquals(aCfg.getEditor(), USE_VIM)
        files = aCfg.getFiles()
        if sys.platform != 'linux2':  return
        self.assertEquals(files, ['../browser/test_Debug.py', '../browser/test_browser.py', 'test_browser.py'])
        self.assertEquals(aCfg.getVerbose(), 0)
                
    # the CLP owns the defaults...
    
        nothing_args = []
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(nothing_args)
        self.assertEquals(aCfg.getEditor(), USE_IDLE)
        files = aCfg.getFiles()
        self.assertEquals(files,  ['Debug.py', 'TreeBox.py', 'alltests.py', 'browser.py', 'listtests.py', 'manytests.py', 'notatest.py', 'piper.py', 'test_Debug.py', 'test_browser.py', ])

    #  in the presence of individual files the other defaults vanish
    
        file_args = ['test_Debug.py']
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(file_args)
        self.assertEquals(aCfg.getEditor(), USE_IDLE)
        files = aCfg.getFiles()
        self.assertEquals(files, ['test_Debug.py'])

    #  in the presence of wildcards, the folder default remains
    
        wild_args = ['many*.py']
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(wild_args)
        self.assertEquals(aCfg.getEditor(), USE_IDLE)
        files = aCfg.getFiles()
        self.assertEquals(files, ['manytests.py'])

        verbose_args = ['--V', 'test_Debug.py']
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(verbose_args)
        self.assertEquals(aCfg.getVerbose(), 1)

    def test_multipleSimilarNodes(self):
        aBrowser = Browser(getFiles(['.']))
        many = aBrowser.myAll.getTestCount()
        
    #  this shows the recursive constructors resist dupes
        
        bBrowser = Browser(getFiles(['.', '.']))
        self.assertEquals(bBrowser.myAll.getTestCount(), many)

    def test_absorbFile(self):
        _putPassingTestInSubFolder()  # assert we >ignore< this!
        assert path.exists('kozmik/__init__.py')
    	
#    	os.system('./browser.py test_browser.py')        
    	
        aBrowser = Browser(['test_browser.py'])
        aDot = aBrowser.myAll.getChild(0)
        
        self.assertEquals(aDot.getName(), '.')
        self.assertEquals(aBrowser.myAll.getChild(1), None)
        aFile = aDot.getChild(0)
        self.assertEquals(aFile.getName(), 'test_browser.py')
        self.assertEquals(aBrowser.myAll.getChild(1), None)
        
    #  this also shows the recursive constructors resist dupes
    
        aBrowser = Browser(['test_browser.py', 'test_browser.py'])
        aDot = aBrowser.myAll.getChild(0)
        self.assertEquals(len(aDot.myChildren), 1)

    def test_absorbFolders(self):
        _putPassingTestInSubFolder()  # assert we >ignore< this!
        assert path.exists('kozmik/__init__.py')
#    	os.system('./browser.py kozmik')
        
        aBrowser = Browser(getFiles(['kozmik']))
        aFolder = aBrowser.myAll.getChild(0)
        
        self.assertEquals(aFolder.getName(), 'kozmik')
        self.assertEquals(aBrowser.myAll.getChild(1), None)
        aDot = aFolder.getChild(0)
        self.assertEquals(aDot.getName(), '.')
        aFile = aDot.getChild(0)
        self.assertEquals(aFile.getName(), _testBullfrog)
        self.assertEquals(aBrowser.myAll.getChild(1), None)

    #  now show that a duped file and folder on the command line
    #  get merged in memory
    
        aBrowser = Browser(getFiles(['.']))
        fileCount = len(aBrowser.myAll.getChild(0).myChildren)

        aBrowser = Browser(getFiles(['.', 'test_Debug.py']))
        nuFileCount = len(aBrowser.myAll.getChild(0).myChildren)
        self.assertEquals(fileCount, nuFileCount)

    def test_absorbWildCards(self):
        aBrowser = Browser(getFiles(['.', 'list*.py']))
        aFolder = aBrowser.myAll.getChild(0)
        fileCount = len(aFolder.myChildren)
        self.assertEquals(fileCount, 1)
        aFile = aFolder.getChild(0)
        self.assertEquals(aFile.getName(), 'listtests.py')        

    def test_absorbPath(self):
        deepFolder, deepFile = _writeDeepFolderPackages()
        aBrowser = Browser(getFiles([deepFile]))
        self._checkDeepThings(aBrowser)

    def test_absorbDeepFolders(self):
        deepFolder, deepFile = _writeDeepFolderPackages()
        aBrowser = Browser(getFiles([deepFolder]))
        self._checkDeepThings(aBrowser)

    def _checkDeepThings(self, aBrowser):
        we_now_expect = [
            '<all>',
                'kozmik',
                    'xylophonist',
                        '.',
                            'chops.py',
                                'TestSuite',
                                    'testFaylur', ]
                                    
        aThing = aBrowser.myAll
        idx = 0
        
        while aThing is not None:
            self.assertEquals(aThing.getTestCount(), 1)

            if aThing.getChild(0) is not None:
                self.assertEquals(len(aThing.myChildren), 1)
                
            self.assertEquals(aThing.getName(), we_now_expect[idx])
            idx += 1
            aThing = aThing.getChild(0)          

        self.assertEquals(idx, len(we_now_expect))

    def test_addChild(self):
        deepFolder, deepFile, deepFile2 = _writeDeepFolderPackagesWithMoreFiles()
#        os.system('xterm')
#        os.system('./browser.py kozmik')
        aBrowser = Browser(getFiles(['kozmik/']))
        count = aBrowser.myAll.getTestCount()
        self.assertEquals(count, 3)

    def test_absorbCommandLine(self):
        deepFolder, deepFile = _writeDeepFolderPackages()
        sys_args = []
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(sys_args)
        
        aBrowser = Browser(aCfg.getFiles())
        self.assertNotEquals(aBrowser.myAll.getChild(1), None)

        aTreeBox = Mock()
        aBrowser.myAll.growTree(aTreeBox)
        calls = aTreeBox.getAllCalls()
        keez = []
        
        for call in calls:
            myKey = call.params[0]
            assert not (myKey in keez)
            keez.append(myKey)
            yek = call.params[2]
            assert yek == '' or yek in keez
            stringCalls = [str(pa.params) for pa in calls]

        for param in (
                 ('/<all>', '<all>', ''),
                 ('/<all>/.', '.', '/<all>'),
                 ('/<all>/./alltests.py', 'alltests.py', '/<all>/.'),
                 ('/<all>/./listtests.py', 'listtests.py', '/<all>/.'),
                 ('/<all>/./listtests.py/TestSuite1/testSlicing', 'testSlicing', '/<all>/./listtests.py/TestSuite1'),
                 ('/<all>/./manytests.py', 'manytests.py', '/<all>/.'),
                 ('/<all>/./manytests.py/test3', 'test3', '/<all>/./manytests.py'),
                 ('/<all>/./test_Debug.py', 'test_Debug.py', '/<all>/.'),
                 ('/<all>/kozmik', 'kozmik', '/<all>'),
                 ('/<all>/kozmik/.', '.', '/<all>/kozmik'),
                 ('/<all>/kozmik/./testBullfrog.py', 'testBullfrog.py', '/<all>/kozmik/.'),
                 ('/<all>/kozmik/./testBullfrog.py/TestSuite', 'TestSuite', '/<all>/kozmik/./testBullfrog.py'),
                 ('/<all>/kozmik/xylophonist', 'xylophonist', '/<all>/kozmik'),
                 ('/<all>/kozmik/xylophonist/.', '.', '/<all>/kozmik/xylophonist'),
                 ('/<all>/kozmik/xylophonist/./chops.py', 'chops.py', '/<all>/kozmik/xylophonist/.'),
                 ('/<all>/kozmik/xylophonist/./chops.py/TestSuite', 'TestSuite', '/<all>/kozmik/xylophonist/./chops.py'),
                 ('/<all>/kozmik/xylophonist/./chops.py/TestSuite/testFaylur', 'testFaylur', '/<all>/kozmik/xylophonist/./chops.py/TestSuite'),
                ):
#            db(param)
            self.assert_(str(param) in stringCalls)

#        os.system('./browser.py')

#    ./browser.py kozmik \*test.py

#  TODO  internationalize these

        for args in (       
                 ['./kozmik/xylophonist'],
                 ['kozmik/xylophonist/'],
                 ['./kozmik/xylophonist/'],
                 ['./kozmik/xylophonist/.'],
                ):
            files = getFiles(args)
            self.assert_(files in ([deepFile], [path.join('.', deepFile)]))
                
        args = ['kozmik/xylophonist']
        files = getFiles(args)
        self.assertEquals(files, [deepFile])
        aBrowser = Browser(files)
        self.assertEquals(aBrowser.myAll.getChild(0).getName(), _subFolder)

    def test_getFiles(self):
        
#  TODO split these up into individual tests

        deepFolder, deepFile = _writeDeepFolderPackages()
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine([])
        files = aCfg.getFiles()
        
        for thing in (
                    ('test_browser.py'), 
                    ('test_Debug.py'),
                    ('manytests.py'), 
                    (deepFile), 
                ):
            self.assert_(thing in files)        

        for aint in (
                    ('.'), 
                    ('kozmik'),
                    ('__init__.py'), 
                    (deepFolder), 
                ):
            self.assert_(not (aint in files)) 

        minimal_args = [deepFile, deepFolder, '*.py', ]
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(minimal_args)
        aFileCount = len(aCfg.getFiles())

        maximal_args = [ deepFile, deepFile,
                         deepFolder, deepFolder,
                         'test*.py', '*.py', 'test*.py', ]
                         
        bCfg = CommandLineParser()
        bCfg.absorbCommandLine(maximal_args)
        bFileCount = len(bCfg.getFiles())

        self.assertEquals(aFileCount, bFileCount)
        
        aBrowser = Browser(aCfg.getFiles())

    #  previous devious calls should result in only one (1) test
    
        self._checkDeepThings(aBrowser)
        
    #  Now show that deep traversal still works by default
        
        aCfg = CommandLineParser()
        aCfg.absorbCommandLine([])
        aBrowser = Browser(aCfg.getFiles())
        self.assertEquals(aBrowser.myAll.getChild(0).getName(), '.')
        
    #  Now show we can tweezer out one nearby and one
    #  deep folder without pulling all the others in

        aCfg = CommandLineParser()
        aCfg.absorbCommandLine(['test_Debug.py', deepFolder, ])
        self.assertNotEquals(aCfg.myWildCards, [])

        aBrowser = Browser(aCfg.getFiles())
#        _printTree(aBrowser.myAll)
        self.assertEquals(aBrowser.myAll.getTestCount(), 3)

    def test_bogusFileOnCommandLine(self):       
        files = ['Debug.py', 'test_Debug.py']
        aBrowser = Browser(files)
        self.assertEquals(aBrowser.myAll.getTestCount(), 2)
        aBrowser = Browser(getFiles(files))
        self.assertEquals(aBrowser.myAll.getTestCount(), 2)

    def test_pyCheck(self):
        try:
            import pychecker
        except ImportError:
            return

        sample_file = 'listtests.py'
        aBrowser = Browser(getFiles([sample_file]))
        aBrowser.myBar = Mock()
        aBrowser.myErrorList = Mock()
        aFile = aBrowser.myAll.getChild(0).getChild(0)
        assert aFile.getName() == sample_file
        got = aFile.pyCheck()
        self.assertEquals(got, 256)  #  picky picky
        calls = aBrowser.myErrorList.getAllCalls()
        call = calls[0]
        self.assertEquals(str(call), 'clear()')

    #  Warnings... expect new versions to bust this test hyper

        expect = [
            'Warnings...',
            '/usr/lib/python2.2/unittest.py:215: Local variable (e) not used',
            'listtests.py:43: Statement appears to have no effect',
        ]
        calls = calls[1:]
        idx = 0
        
        for x in xrange(len(expect) * 2):
            x = x
            call = calls[idx]
            shouldBe = expect[idx]
            if call.getName() == 'update':  continue
            self.assertEquals(call.getName(), 'insert')
            self.assertEquals(len(call.params), 2)
            self.assertEquals(call.getParam(0), 'end')
            self.assertEquals(call.getParam(1), shouldBe)
            idx += 1

        self.assertEquals(len(calls), len(expect) * 2)
        calls = aBrowser.myBar.getAllCalls()
        self.assertEquals(calls[0].getName(), 'setMessage')
        self.assertEquals(calls[1].getName(), 'setMessage')

    def test_getAllFileNames(self):
        todo = _writeFileWithTodoLine()
        
        argv = ['listtests.py', _test_orangeHerring]
        aBrowser = Browser(getFiles(argv))
        aFolder = aBrowser.myAll.getChild(0)
        aFile = aFolder.getChild(0)
        aSuite = aFile.getChild(0)
        aTest = aSuite.getChild(0)
        self.assertEquals(aTest.getAllFileNames(), argv[0:1])
        self.assertEquals(aSuite.getAllFileNames(), argv[0:1])
        self.assertEquals(aFile.getAllFileNames(), argv[0:1])
        self.assertEquals(aFolder.getAllFileNames(), argv)
        self.assertEquals(aBrowser.myAll.getAllFileNames(), argv)
        argv = ['test_browser.py']
        expect = argv + ['browser.py']
        aBrowser = Browser(getFiles(argv))
        aFolder = aBrowser.myAll.getChild(0)
        aFile = aFolder.getChild(0)
        aSuite = aFile.getChild(0)
        aTest = aSuite.getChild(0)
#        self.assertEquals(aTest.getAllFileNames(), expect)
# TODO:  self.assertEquals(aSuite.getAllFileNames(), argv)
#        self.assertEquals(aFile.getAllFileNames(), expect)
#        self.assertEquals(aFolder.getAllFileNames(), expect)
#        self.assertEquals(aBrowser.myAll.getAllFileNames(), expect)
        
    def test_grep(self):
        todo = _writeFileWithTodoLine()
        
        argv = ['listtests.py', _test_orangeHerring]
        aBrowser = Browser(getFiles(argv))
        aFolder = aBrowser.myAll.getChild(0)
        aFile = aFolder.getChild(1)
        aSuite = aFile.getChild(0)
        aTest = aSuite.getChild(0)

        for thing in (aTest, aSuite, aFile, aFolder, aBrowser.myAll):
            aBrowser.myErrorList = Mock()
            aBrowser.myBar = Mock()
            if sys.platform != 'linux2':  return
            thing.grep('TODO')
            
  # TODO oaoo the calls to test what's in a list box...
            
            calls = aBrowser.myErrorList.getAllCalls()
            self.assertEquals(len(calls), 3)
            zeroth = str(calls[0])
            self.assertEquals(zeroth, 'clear()')
            first = str(calls[1])
            line = "insert('end', '%s:6:    # %s')" % (_test_orangeHerring, todo)
            self.assertEquals(first, line)
            second = str(calls[2])
            self.assertEquals(second, 'update()')     

        return

    def tearDown(self):
        for killit in (_test_orangeHerring, _subFolder,):
            removeForcedRecursive(killit)


def _writeFileWithTodoLine():
    todo = 'TODO subvert reality'
    line = 'pass\n    # %s\n        pass' % todo
        
    open(_test_orangeHerring, 'w') \
        .write(_instaTest % line)
            
    return todo
    
    
def _printTree(aThing, depth = 0):
    print '  ' * depth, aThing.getName()
    for kid in aThing.myChildren:
        _printTree(kid, depth + 1)


_subFolder = 'kozmik'
_testBullfrog = 'testBullfrog.py'
_subFile = path.join(_subFolder, _testBullfrog)
_subInit = path.join(_subFolder, '__init__.py')

_instaSuite = '''def suite():
    suite = unittest.makeSuite(RedHerringTestCase)
    return unittest.TestSuite((suite,))
'''

_instaTest = '''
import unittest, sys
class RedHerringTestCase(unittest.TestCase):
    def testFaylur(self):
        %s
''' + _instaSuite


def _putPassingTestInSubFolder(withInit = 1):
    os.mkdir(_subFolder)
    open(_subFile, 'w').write(_instaTest % 'pass')
    if not withInit:  return 
    open(_subInit, 'w').write('#stub')


def _writeDeepFolderPackages():
    _putPassingTestInSubFolder()
    deepFolder = path.join(_subFolder, 'xylophonist')
    os.mkdir(deepFolder)
    deepFile = path.join(deepFolder, 'chops.py')
    deepInit = path.join(deepFolder, '__init__.py')
    file(deepInit, 'w').write('#stub')
    file(deepFile, 'w').write(_instaTest % 'pass')
    return deepFolder, deepFile


def _writeDeepFolderPackagesWithMoreFiles():
    deepFolder, deepFile = _writeDeepFolderPackages()
    deepFile2 = path.join(deepFolder, 'Impeccability.py')
    file(deepFile2, 'w').write(_instaTest % 'pass')
    return deepFolder, deepFile, deepFile2


def _mockProgress(aNode):
    result = aNode._makeResult()
    bar = Mock()
    lister = Mock ()
    result.setUserView(bar, lister, aNode.getTestCount())
    aNode.callTests(result)
    return bar, lister


_test_orangeHerring = 'test_orangeHerring.py'


def _makeCensoredBrowser():
    files = ['alltests.py', 'listtests.py', 'notatest.py', 'manytests.py',]
    aBrowser = Browser(files)
    aFolder = aBrowser.myAll.getChild(0)
    assert aFolder.getName() == '.'
      #  censors test_browser.py to prevent infinite recursion ;-)
      #  also prevent subsequent test count numbers from
      #  changing at add test time
    return aBrowser


def touch(filename, new_time = time.time()) :
    os.utime(filename, (new_time, new_time))


def getFiles(paths):
    aCfg = CommandLineParser()
    aCfg.absorbCommandLine(paths)
    return aCfg.getFiles()


def suite():
    suite1 = unittest.makeSuite(BrowserTestCase)
    return unittest.TestSuite((suite1,))


if __name__ == '__main__':
    # When this module is executed from the command-line, run all its tests
    unittest.main()

